import argparse
import functools
import os
import os.path
import pickle
import urllib
import urllib.parse
import zipfile
from urllib.request import urlretrieve

import requests
import requests.exceptions
from bs4 import BeautifulSoup


def parse_args(*args):
    parser = argparse.ArgumentParser(description="Mirror simfiles from ZIV")

    parser.add_argument('categories', type=str, nargs='*', help='ZIV category pages to mirror',
                        default=['https://zenius-i-vanisher.com/v5.2/simfiles.php?category=latest20official'])

    parser.add_argument('--songdir', type=str, help="Directory to keep songs in", default="songs")
    parser.add_argument('--zipdir', type=str, help="Directory to keep downloaded zip files in", default="zips")

    feature = parser.add_mutually_exclusive_group(required=False)
    feature.add_argument('--dry-run', '-n',
                         help="Only perform a dry run; don't send any pings",
                         dest='dry_run', action='store_true')
    feature.add_argument('--no-dry-run',
                         help="Send pings normally",
                         dest='dry_run', action='store_false')
    feature.set_defaults(dry_run=False)

    return parser.parse_args(*args)


def retrieve(url, filename, save_headers=None, extract=None, **kwargs):
    print(f'Downloading {url} -> {filename}')
    try:
        req = requests.get(url, **kwargs, stream=True)
        if req.status_code == 200:
            with open(filename, 'wb') as output:
                for chunk in req.iter_content(1024):
                    output.write(chunk)

            if extract:
                with zipfile.ZipFile(filename, 'r') as zip:
                    print(f'Extracting into {extract}')
                    os.makedirs(extract, exist_ok=True)
                    zip.extractall(extract)

            if save_headers:
                pickle.dump(req.headers, open(save_headers, 'wb'))
        elif req.status_code == 304:
            print("Not modified")
        else:
            print(f"Error: {req.status_code} {req.text}")
    except requests.exceptions.BaseHTTPError as e:
        print(f'Error downloading: {e.msg}')
    except zipfile.BadZipFile:
        print(f'Not a zip file: {filename}')
    except KeyboardInterrupt as e:
        print(f'Download aborting...')
        if os.path.isfile(filename):
            print(f'Removing partial file {filename}')
            os.unlink(filename)
        raise e

    return req.headers


@functools.lru_cache()
def get_page(cat_url):
    request = requests.get(cat_url)
    return BeautifulSoup(request.text, features="html.parser")


def load_prev_headers(filename, header_file):
    req_headers = {}
    if os.path.isfile(header_file) and os.path.isfile(filename):
        prev_headers = pickle.load(open(header_file, 'rb'))
        if 'etag' in prev_headers:
            req_headers['If-None-Match'] = prev_headers['etag']
        if 'last-modified' in prev_headers:
            req_headers['If-Modified-Since'] = prev_headers['last-modified']
    return req_headers


def mirror(cat_url, args):
    page = get_page(cat_url)

    group_urls = {}

    if 'viewsimfilecategory.php' in cat_url:
        simgroup = page.find('div', {'class': 'headertop'}).h1
        group_url = cat_url
    else:
        simgroup = None

    for row in page.find_all('tr'):
        simfile = row.find(
            "a", href=lambda href: href and "viewsimfile.php" in href)
        group_link = row.find(
            "a", href=lambda href: href and "viewsimfilecategory.php" in href)
        if group_link:
            simgroup = group_link
            group_url = group_link['href']

        if not (simfile and simgroup):
            continue

        songname = ' '.join(simfile.get_text().replace('/', '-').split())
        groupname = ' '.join(simgroup.get_text().replace('/', '-').split())

        print(f"collection: '{groupname}' simfile: '{songname}'")

        simlink = simfile['href']
        try:
            sim_id = urllib.parse.parse_qs(urllib.parse.urlparse(
                simfile['href']).query)['simfileid'][0]
        except KeyError:
            print(f"WARNING: no simfileid found on URL {simlink}")
            continue

        group_urls[groupname] = urllib.parse.urljoin(cat_url, group_url)

        url = f'https://zenius-i-vanisher.com/v5.2/download.php?type=ddrsimfile&simfileid={sim_id}'

        if args.dry_run:
            print(f"Dry run requested, not downloading {url}")
            continue

        filename = os.path.join(args.zipdir, f'{sim_id}.zip')
        headers = os.path.join(args.zipdir, f'{sim_id}.headers')
        req_headers = load_prev_headers(filename, headers)

        retrieve(url, filename, extract=os.path.join(args.songdir, groupname),
                 headers=req_headers, save_headers=headers)

    for groupname, group_url in group_urls.items():
        page = get_page(group_url)
        banner_urls = {urllib.parse.urljoin(group_url, banner['src'])
                       for banner in page.select('p.centre img')
                       if 'simfileNoBanner.png' not in banner['src']}
        for url in banner_urls:
            filename = os.path.join(args.songdir, groupname, 'banner.png')
            headers = os.path.join(args.zipdir, f'{groupname}-banner.headers')
            req_headers = load_prev_headers(filename, headers)
            retrieve(url, filename, extract=False,
                     headers=req_headers, save_headers=headers)


if __name__ == "__main__":
    args = parse_args()

    os.makedirs(args.songdir, exist_ok=True)
    os.makedirs(args.zipdir, exist_ok=True)

    for url in args.categories:
        mirror(url, args)
